#ifndef AST_H
#define AST_H

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <deque>
#include <set>
#include "AstVisitor.h"

using namespace std;

class Ast;
class FctDef;
class Block;
class Statement;
class Assignment;
class VarDecl;




class Parameter {
 public:
  string paramType;
  string param;
};


class FctDef {

 protected:

  string qualifiers;

  string returnType;
  string fctName;

  deque<Parameter*> params;

  Block *bdy;


 public:

  virtual void accept(AstVisitor& v) {
    v.visit(*this);
  }

  string getFctName() { return fctName; }
  
  Block *getBody() { return bdy; }

  deque<Parameter*>& getParams() { return params; }

};


class Statement {

 public:

  Statement() { }
  virtual ~Statement() { }
  virtual void accept(AstVisitor& v) { }

};

class Block {

 protected:

  deque<Statement*> stmts;

 public:

  virtual void accept(AstVisitor& v) {
    v.visit(*this);
  }

  void addStmt(Statement* s) {
    stmts.push_back(s);
  }

  deque<Statement*>& getStmts() { return stmts; }

};

class Expression {

 public:

  Expression() { }
  ~Expression() { }

  virtual bool isAtom() { return false; }
  virtual bool isLiteral() { return false; }
  virtual bool isIdentifier() { return false; }

  virtual void accept(AstVisitor& v) { }

  virtual string toString() const {
    return string("<undefined expression>");
  }

};


class NaryExpression : public Expression {

 protected:

  string fct;

  deque<Expression*> operands;

 public:

  NaryExpression(string f) { fct = f; }

  void addOperand(Expression *op) {
    operands.push_back(op);
  }

  deque<Expression*>& getOperands() { return operands; }

  string getFct() { return fct; }

  virtual void accept(AstVisitor& v) {
    v.visit(*this);
  }

  virtual string toString() const { 

    string s(fct);

    s += "(";

    bool first = true;
    for ( deque<Expression*>::const_iterator
	    e = operands.begin();
	  e != operands.end();
	  e++ ) {
      if ( not first ) {
	s += ",";
      }
      first = false;
      s += (*e)->toString();
    }

    s += ")";
    return s;

  }


};






class Literal : public Expression {

  string theAtom;

 public:

  Literal(string s) { theAtom = s; }

  virtual bool isAtom() { return true; }
  virtual bool isLiteral() { return true; }
  virtual string getAtom() { return theAtom; }


  virtual void accept(AstVisitor& v) {
    v.visit(*this);
  }

  virtual string toString() const { return theAtom; }

};


class Identifier : public Expression {

  string theAtom;

 public:

  Identifier(string x) { theAtom = x; }
  virtual bool isAtom() { return true; }
  virtual bool isIdentifier() { return true; }
  virtual string getAtom()  { return theAtom; }


  virtual void accept(AstVisitor& v) {
    v.visit(*this);
  }

  virtual string toString() const { return theAtom; }


};



class IfElseStatement : public Statement {

  Expression *condition;

  Block *ifBlk;
  Block *elseBlk;
  

 public: 

  IfElseStatement(Expression* c, Block* b1, Block* b2) {
    condition = c;
    ifBlk = b1;
    elseBlk = b2;
  }

  virtual void accept(AstVisitor& v) {
    v.visit(*this);
  }

  Expression *getCnd() { return condition; }
  Block* getIfBlk() { return ifBlk; }
  Block* getElseBlok() { return elseBlk; }

};


class IfStatement : public Statement {

  Expression *condition;

  Block *ifBlk;
  

 public: 

  IfStatement(Expression* c, Block* b1) {
    condition = c;
    ifBlk = b1;
  }

  virtual void accept(AstVisitor& v) {
    v.visit(*this);
  }

  Expression *getCnd() { return condition; }
  Block* getIfBlk() { return ifBlk; }

};


class Assignment : public Statement {

  Expression *lhs;
  Expression *rhs;

 public:

  Assignment(Expression *l, Expression *r) {
    lhs = l; rhs = r;
  }

  Expression* getLhs() { return lhs; }
  Expression* getRhs() { return rhs; }

  virtual void accept(AstVisitor& v) {
    v.visit(*this);
  }

  string toString() const {
    string s = lhs->toString() + " = " + rhs->toString() + ";";
    return s;
  }

};


class WhileStatement : public Statement {

 protected:
  Expression *cnd;
  Block *blk;

 public:

  WhileStatement(Expression *c, Block *b) {
    cnd = c;
    blk = b;
  }

  Expression *getCnd() { return cnd; }
  Block *getBlk() { return blk; }

  virtual void accept(AstVisitor& v) {
    v.visit(*this);
  }

};



class VarDecl :  public Statement {

 protected:

  string theType;
  string theVar;

 public:

  VarDecl(string t, string v) {
    theType = t;
    theVar = v;
  }


  virtual void accept(AstVisitor& v) {
    v.visit(*this);
  }

};


#endif
